from torchvision import transforms
import yaml
import torch.nn as nn
import torch

class EfficientNetWithFeatures(nn.Module):
    """
    Universal wrapper for EfficientNet models that outputs both classification and feature maps
    Works with: B0, B1, B2, B3, B4, B5, B6, B7, V2_S, V2_M, V2_L
    Perfect for Faster-CAM in C#!

    This wrapper automatically detects the model architecture and wraps it to output:
    - logits: Classification output (B, num_classes)
    - features: Feature maps from last layer (B, C, H, W)
    """
    def __init__(self, model):
        super().__init__()

        # Store the original model components
        self.features = model.features
        self.avgpool = model.avgpool
        self.classifier = model.classifier

        # Detect model type for debugging
        model_name = model.__class__.__name__
        print(f"Wrapped model: {model_name}")

    def forward(self, x):
        # Extract features from last convolutional layer
        features = self.features(x)  # Shape: (B, C, H, W)

        # Classification path
        x = self.avgpool(features)
        x = torch.flatten(x, 1)
        logits = self.classifier(x)  # Shape: (B, num_classes)

        # Return both: classification logits AND feature maps
        return logits, features


class Utils:

    def PrepareTrain_Augmentation(settingParams, TrainingParams):
            
        transform_list = [transforms.Resize(((int)(settingParams['ImageSize']), (int)(settingParams['ImageSize'])))]
        
        if TrainingParams['FlipLR']['Value'] > 0:
            transform_list.append(transforms.RandomHorizontalFlip(p=TrainingParams['FlipLR']['Value']))  # 50% chance

        if TrainingParams['FlipUD']['Value'] > 0:
            transform_list.append(transforms.RandomVerticalFlip(p=TrainingParams['FlipUD']['Value']))  # 0% chance

        if TrainingParams['Degrees']['Value'] > 0:
            transform_list.append(transforms.RandomRotation(degrees=TrainingParams['Degrees']['Value']))  # 0 degrees

        if TrainingParams['Translate']['Value'] > 0:
            translate = TrainingParams['Translate']['Value']
            transform_list.append(transforms.RandomAffine(degrees=0, translate=(translate, translate)))  # Translate by 0.1

        if TrainingParams['Scale']['Value'] > 0:
            scale = TrainingParams['Scale']['Value']
            transform_list.append(transforms.RandomAffine(degrees=0, scale=(1-scale, 1+scale)))  # Scale by ±0.5

        if TrainingParams['Shear']['Value'] > 0:
            transform_list.append(transforms.RandomAffine(degrees=0, shear=TrainingParams['Shear']['Value']))  # 0 shear

        if TrainingParams['Perspective']['Value'] > 0:
            transform_list.append(transforms.RandomPerspective(distortion_scale=TrainingParams['Perspective']['Value'], p=0.5))  # 0 perspective

        if TrainingParams['Erasing']['Value'] > 0:
            transform_list.append(transforms.RandomErasing(p=TrainingParams['Erasing']['Value'], scale=(0.02, 0.33), ratio=(0.3, 3.3)))  # 40% chance

        
        if( TrainingParams["HsvH"]['Value'] > 0 or  
           TrainingParams["HsvS"]['Value'] > 0):
            
            transform_list.append(transforms.ColorJitter(hue=TrainingParams["HsvH"]['Value'],
                                                         saturation=TrainingParams["HsvS"]['Value']))
        
        
        # # HSV adjustments (torchvision doesn't support direct HSV, so use ColorJitter)
        # if any(TrainingParams[key]['Value'] > 0 for key in ['Hsv_h', 'Hsv_s', 'Hsv_v']):
        #     transform_list.append(transforms.ColorJitter(
        #         hue=TrainingParams['Hsv_h']['Value'],  # Hue shift
        #         saturation=TrainingParams['Hsv_s']['Value'],  # Saturation shift
        #         brightness=TrainingParams['Hsv_v']['Value']  # Value/Brightness shift
        #     ))


        Train_Transform = transforms.Compose(transform_list)

        return Train_Transform

    def load_yaml(file_path):
        """Load and parse a YAML file."""
        with open(file_path, 'r') as file:
            return yaml.safe_load(file)
    
    def get_input_shape(model, default_hw=(224, 224)):
        """
        Extract the input shape of a PyTorch model as a tuple of numbers.
        Args:
            model: PyTorch model (nn.Module)
            default_hw: Tuple of (height, width) for CNNs if not specified (default: (224, 224))
        Returns:
            Tuple of integers representing input shape (excluding batch dimension)
        """
        # Get the first layer
        for module in model.modules():
            if isinstance(module, (nn.Conv2d, nn.Linear, nn.LSTM, nn.GRU)):
                first_layer = module
                break
        else:
            raise ValueError("No recognizable layer (Conv2d, Linear, LSTM, GRU) found in model")

        if isinstance(first_layer, nn.Conv2d):
            # For CNNs: (channels, height, width)
            return (first_layer.in_channels,) + default_hw
        elif isinstance(first_layer, nn.Linear):
            # For linear layers: (input_features,)
            return (first_layer.in_features,)
        elif isinstance(first_layer, (nn.LSTM, nn.GRU)):
            # For RNNs: (sequence_length, input_size)
            # sequence_length is often variable, so we return input_size
            return (None, first_layer.input_size)  # None for variable sequence length
        else:
            raise ValueError("Unsupported layer type")


"""
MemoLib - A Python library for machine learning model training and inference
"""

try:
    from importlib.metadata import version, metadata

    __version__ = version("MemoLib")
    _meta = metadata("MemoLib")
    __description__ = _meta["Summary"]
    __license__ = _meta["License"]
    __author__ = _meta["Author"]
    __email__ = _meta["Author-email"]
except Exception:
    # Fallback values if package is not installed
    __version__ = "0.1.7"
    __description__ = "A Python library for machine learning model training and inference"
    __license__ = "MIT"
    __author__ = "NghiaPham"
    __email__ = "nghiaphamkthp2401@gmail.com"

from .Model import MemoModel, IModel, eExportType, MemoResult, MemoPrediction, ModelConfig
from .Model.eModel import eModelType, eDetectionModel, eClassifyModel, eModelTask
from .DataSerializer import *
from .DatasetFormatConvert import *
from .Loss import *

__all__ = [
    "MemoModel",
    "IModel",
    "eExportType",
    "MemoResult",
    "MemoPrediction",
    "ModelConfig",
    "eModelType",
    "eDetectionModel",
    "eClassifyModel",
    "eModelTask"
]



import torch
import torch.nn as nn
import torch.nn.functional as F

class FocalLoss(nn.Module):
    def __init__(self, alpha=0.25, gamma=2.0, reduction='mean'):
        super(FocalLoss, self).__init__()
        self.alpha = alpha
        self.gamma = gamma
        self.reduction = reduction

    def forward(self, inputs, targets):
        ce_loss = nn.CrossEntropyLoss(reduction='none')(inputs, targets)
        pt = torch.exp(-ce_loss)
        focal_loss = self.alpha * (1 - pt) ** self.gamma * ce_loss
        if self.reduction == 'mean':
            return focal_loss.mean()
        elif self.reduction == 'sum':
            return focal_loss.sum()
        return focal_loss

class StandardCrossEntropy:
    def __init__(self):
        self.criterion = nn.CrossEntropyLoss()
    
    def __call__(self, inputs, targets):
        return self.criterion(inputs, targets)

class BinaryCrossEntropy:
    def __init__(self, weight=None):
        self.criterion = nn.BCEWithLogitsLoss(weight=weight)
    
    def __call__(self, logits, targets):
        # Convert logits [batch, 2] → single output [batch, 1]
        # Take logits for positive class (class 1)
        binary_logits = logits[:, 1] - logits[:, 0]  # Log odds ratio
        targets_float = targets.float()
        return self.criterion(binary_logits, targets_float)
    
class LabelSmoothingCrossEntropy(nn.Module):
    def __init__(self, smoothing=0.1):
        super().__init__()
        self.smoothing = smoothing
        
    def forward(self, inputs, targets):
        log_prob = F.log_softmax(inputs, dim=-1)
        nll_loss = -log_prob.gather(dim=-1, index=targets.unsqueeze(1))
        nll_loss = nll_loss.squeeze(1)
        smooth_loss = -log_prob.mean(dim=-1)
        loss = (1.0 - self.smoothing) * nll_loss + self.smoothing * smooth_loss
        return loss.mean()

class MarginLoss(nn.Module):
    def __init__(self, margin=1.0):
        super().__init__()
        self.margin = margin
        
    def forward(self, inputs, targets):
        # Convert to binary problem: positive vs negative scores
        positive_scores = inputs[range(len(targets)), targets]
        negative_scores = inputs[range(len(targets)), 1 - targets]  # For binary case
        
        losses = F.relu(self.margin - positive_scores + negative_scores)
        return losses.mean()

class DiceLoss(nn.Module):
    def __init__(self, smooth=1e-6):
        super().__init__()
        self.smooth = smooth
        
    def forward(self, inputs, targets):
        probs = F.softmax(inputs, dim=1)
        targets_one_hot = F.one_hot(targets, num_classes=inputs.size(1)).float()
        
        intersection = (probs * targets_one_hot).sum()
        union = probs.sum() + targets_one_hot.sum()
        
        dice = (2.0 * intersection + self.smooth) / (union + self.smooth)
        return 1 - dice


from os import listdir, path, makedirs
from shutil import copy

imgExtension = [".jpeg", ".jpg", ".png", ".bmp"]


def Yolo2Classify(datasetDir: str, exportDir: str):
    
    errorCount = 0
    errorList = []

    imagesDir = path.join(datasetDir, "images")
    labelsDir = path.join(datasetDir, "labels")

    processedCount = 0
    totalLabel = len(listdir(labelsDir))
    print(f"Found {totalLabel} files.")

    if(path.exists(imagesDir) == False):
        raise Exception("Images directory does not exits in dataset directory.")
    if(path.exists(labelsDir) == False):
        raise Exception("Labels directory does not exits in dataset directory.")
    if(path.exists(exportDir == False)):
        print("Export directory does not exits, try create new one!")
        makedirs(exportDir, exist_ok= True)
    
    for labelFileName in listdir(labelsDir):
        
        imgFilePath = ""
        for ext in imgExtension:
            
            if( path.exists( path.join(imagesDir, labelFileName[:-4] + ext))):
                imgFilePath = path.join(imagesDir, labelFileName[:-4] + ext)
                break
        
        if(imgFilePath == ""):
            errorCount += 1
            errorList.append(labelFileName)
            print(f"Image path: {imgFilePath} not found!")

        else:
            
            imgFileName = path.basename(imgFilePath)
            rawImgPath = path.join(imagesDir, imgFileName)
            with open(path.join(labelsDir, labelFileName)) as file:

                lines = file.readlines()
                
                if(len(lines) == 0):
                    exportImageDir = path.join(exportDir, "Background")
                    exportImagePath = path.join(exportImageDir, imgFileName)

                    if(path.exists(exportImageDir == False)):
                        makedirs(exportImageDir, exist_ok= True)

                    copy(rawImgPath, exportImagePath)

                else:

                    try:

                        for line in lines:
                            
                            indexLabel = line.split(' ')[0]
                            exportImageDir = path.join(exportDir, indexLabel)
                            exportImagePath = path.join(exportImageDir, imgFileName)

                            if(path.exists(exportImageDir == False)):
                                makedirs(exportImageDir, exist_ok= True)

                            if(path.exists(exportImagePath)):
                                continue
                            else:
                                copy(rawImgPath, exportImagePath)
                    
                    except Exception as ex:
                        print(f"Error in extract label format: {labelFileName}. Msg: {ex}")
                        errorCount +=1
                        errorList.append(labelFileName)
        
        processedCount +=1
        print(f"Processed {processedCount}/{totalLabel}")
    print(f"Process Done! Error Count: {errorCount}")
    print(f"Label Error: ")
    if(errorCount > 0 ):
        for labelError in errorList :
            print(labelError)
        